"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const FollowRedirects = require("follow-redirects");
const Http = require("http");
const Https = require("https");
const url_1 = require("url");
const path_1 = require("path");
const Debug = require("debug");
const utils_1 = require("./utils");
const NativeAgents = { http: Http, https: Https };
const logger = Debug('acp:service:proxy');
class Proxy {
    constructor(options) {
        this.options = options;
        this.isSecure = false;
        this.isSecure = Proxy.isSSL.test(options.target);
        this.target = url_1.parse(this.options.target);
        this.port = this.isSecure ? 443 : 80;
        const agents = (this.options.followRedirects ? FollowRedirects : NativeAgents);
        this.agent = this.isSecure ? agents.https : agents.http;
    }
    request(path, method, headers, dataStream) {
        logger(`Request(${method}) ${path}`);
        const proxyCtx = {
            pipes,
            toPromise,
            prxoy: this,
        };
        try {
            const options = this.getRequestOptions(path, method, headers);
            const request = this.agent.request(options);
            if (this.options.timeout) {
                request.setTimeout(this.options.timeout, () => request.abort());
            }
            if (dataStream && dataStream.pipe) {
                dataStream.pipe(request);
            }
            request.end();
            return new Promise((resolve, reject) => {
                request.on('response', (response) => {
                    resolve({ ...proxyCtx, request, response });
                });
                request.on('error', (e) => resolve({ ...proxyCtx, request, response: { ...Proxy.ErrorResponse }, message: e.message }));
                request.on('abort', (e) => resolve({ ...proxyCtx, request, response: { ...Proxy.ErrorResponse }, message: e.message }));
            });
        }
        catch (e) {
            return Promise.resolve({
                ...proxyCtx,
                message: e.message,
                response: { ...Proxy.ErrorResponse }
            });
        }
    }
    getRequestOptions(path, method, headers) {
        const options = {
            port: this.target.port || this.port,
            method,
            headers: { ...headers },
            host: this.target.host,
            path: path_1.join(this.target.path, path)
        };
        if (typeof options.headers.connection !== 'string' || !Proxy.upgradeHeader.test(options.headers.connection)) {
            options.headers.connection = 'close';
        }
        if (this.options.changeOrigin) {
            options.headers.host = options.headers.hostname = this.target.host;
        }
        if ((method === 'DELETE' || method === 'OPTIONS')
            && !headers['content-length']) {
            headers['content-length'] = '0';
            delete headers['transfer-encoding'];
        }
        return options;
    }
}
Proxy.isSSL = /^https/;
Proxy.upgradeHeader = /(^|,)\s*upgrade\s*($|,)/i;
Proxy.ErrorResponse = { statusCode: 503, statusMessage: 'Service Unavailable', headers: {} };
function createProxy(options) {
    const proxy = new Proxy(options);
    return (path, method, headers, dataStream) => proxy.request(path, method, headers, dataStream);
}
exports.createProxy = createProxy;
function pipes(ctx) {
    const { response, message } = this;
    const { statusCode, statusMessage, headers } = response;
    ctx.statusCode = statusCode;
    ctx.statusMessage = statusMessage;
    ctx.state.headers = utils_1.writeHeaders(headers, ctx);
    ctx.body = message || response;
    return this;
}
function toPromise(ctx) {
    const proxyCtx = this;
    const { response, message } = proxyCtx;
    return new Promise((resolve, reject) => {
        const { statusCode, statusMessage } = response;
        const ob = {
            data: '',
            statusCode,
            statusMessage,
            headers: ctx.state.headers || utils_1.writeHeaders(response.headers),
            ...proxyCtx,
            toJSON() {
                return {
                    data: this.data,
                    statusCode: this.statusCode,
                    statusMessage: this.statusMessage,
                    headers: this.headers
                };
            },
            message: ''
        };
        if (!response.on) {
            return reject({ ...ob, message });
        }
        response.on('error', (e) => reject({ ...ob, message: e.message }));
        response.on('data', (chunk) => ob.data += chunk);
        response.on('end', () => {
            if (ob.statusCode >= 200 && ob.statusCode < 300) {
                return resolve(ob);
            }
            reject({ ...ob, message });
        });
    });
}
